/* $Id: deattack.c,v 1.1 1998/06/11 00:09:35 kivinen Exp $
 * Cryptographic attack detector for ssh - source code 
 * (C)1998 CORE-SDI, Buenos Aires Argentina
 * Ariel Futoransky(futo@core-sdi.com)
 */


#include "includes.h"
#include "deattack.h"
#include "getput.h"
#include "xmalloc.h"

/* Hashing constants */
#define HASH_MINSIZE (8 * 1024)
#define HASH_ENTRYSIZE (2)
#define HASH_FACTOR (2)
#define HASH_UNUSEDCHAR (0xff)
#define HASH_UNUSED (0xffff)
#define HASH_IV     (0xfffe)

#define HASH_MINBLOCKS  (8)

/* SSH Constants */
#define SSH_MAXBLOCKS (32 * 1024)
#define SSH_BLOCKSIZE (8)
 
/* Hash function (Input keys are cipher results) */
#define HASH(x) GET_32BIT(x)


/* 
   htocrcl (host to crc long)
   converts hosts byte ordering to crc byte ordering 
   */
word32 htocrcl(word32 t)
{
  unsigned char *c, a;
  
  t = htonl(t);
  c = (unsigned char *) &t;
  
  a = c[0]; c[0] = c[3]; c[3] = a;
  a = c[1]; c[1] = c[2]; c[2] = a;
  
  return (t);
}

/*
  check_crc 
  detects if a block is used in a particular pattern 
  */ 
int check_crc(unsigned char *S, unsigned char *buf,
	      word32 len, unsigned char *IV)
{
  word32 crc1, crc2, crc;
  unsigned char *c;
  unsigned char s[SSH_BLOCKSIZE];
  
  memcpy(s, S, sizeof(s));
  
  crc1 = crc32(buf, len);
  for (c = buf; c < buf + len; c += SSH_BLOCKSIZE)
    {
      if (!memcmp(c, s, SSH_BLOCKSIZE))
	*c ^= 0x80;
    }
  *s ^= 0x80;
  crc2 = crc32(buf, len);
  for (c = buf; c < buf + len; c += SSH_BLOCKSIZE)
    {
      if (!memcmp(c, s, SSH_BLOCKSIZE))
	*c ^= 0x80;
    }
  *s ^= 0x80;
  
  if (crc1 == crc2) return ( 1 );
  
  
  if (IV && (!memcmp(c, IV, SSH_BLOCKSIZE)))
    {
      crc = htocrcl(crc32(&IV, SSH_BLOCKSIZE));
      PUT_32BIT(buf, GET_32BIT(buf) ^ crc);
      
      crc1 = crc32(buf, len);
      for (c = buf; c < buf + len; c += SSH_BLOCKSIZE)
	{
	  if (!memcmp(c, s, SSH_BLOCKSIZE))
	    *c ^= 0x80;
	}
      *s ^= 0x80;
      crc2 = crc32(buf, len);
      for (c = buf; c < buf + len; c += SSH_BLOCKSIZE)
	{
	  if (!memcmp(c, s, SSH_BLOCKSIZE))
	    *c ^= 0x80;
	}
      *s ^= 0x80;
      PUT_32BIT(buf, GET_32BIT(buf) ^ crc);
      if (crc1 == crc2) return ( 1 );
      
    }
  return (0);
}

/*
  detect_attack
  Detects a crc32 compensation attack on a packet
 */
int detect_attack(unsigned char *buf, word32 len, unsigned char *IV)
{
  static word16 *h = (word16 *)NULL;
  static word32 n;
  word32 l, i, j;
  unsigned char  *c, *d;
  
  assert(len <= (SSH_MAXBLOCKS * SSH_BLOCKSIZE));
  assert(len % SSH_BLOCKSIZE == 0);
  
  l = len / SSH_BLOCKSIZE * HASH_FACTOR;
  
  if (l <= HASH_MINBLOCKS)
    {
      for (c = buf; c < buf + len; c += SSH_BLOCKSIZE)  
	{
	  if (IV && (!memcmp(c, IV, SSH_BLOCKSIZE)))
	    {	
	      if ((check_crc(c, buf, len, IV)))
		return(DEATTACK_DETECTED);
	    }
	  
	  for (d = buf; d < c ; d += SSH_BLOCKSIZE)
	    {
	      if (!memcmp(c, d, SSH_BLOCKSIZE))
		{
		  if ((check_crc(c, buf, len, IV)))
		    return(DEATTACK_DETECTED);
		}
	    }
	}
      return(DEATTACK_OK);
    }
  
  if (l < (HASH_MINSIZE / SSH_BLOCKSIZE)) l = HASH_MINSIZE / SSH_BLOCKSIZE;
  
  if (h == NULL)
    {
      n = l;
      h = (word16 *)xmalloc(n * HASH_ENTRYSIZE);
    }
  else if (l > n)
    {
      n = l;
      h = (word16 *)xrealloc(h, n * HASH_ENTRYSIZE);
    }
  memset(h, HASH_UNUSEDCHAR , n * HASH_ENTRYSIZE);
  
  if (IV)	
    h[HASH(IV) % n] = HASH_IV;
  
  for (c = buf, j = 0; c < (buf + len); c += SSH_BLOCKSIZE, j++)
    {
      for (i = HASH(c) % n; h[i] != HASH_UNUSED; i = (i + 1 < n) ? i + 1 : 0)
	{
	  if (h[i] == HASH_IV) 
	    {
	      if ((!memcmp(c, IV, SSH_BLOCKSIZE)) && 
		  (check_crc(c, buf, len, IV)))
		return(DEATTACK_DETECTED); 
	    } else
	      if ((!memcmp(c, buf + h[i] * SSH_BLOCKSIZE, SSH_BLOCKSIZE)) && 
		  (check_crc(c, buf, len, IV)))
		return(DEATTACK_DETECTED);
	}
      h[i] = j;
    }
  return (DEATTACK_OK);
}

